
#ifndef PLS_TASK_MANAGER_H_
#define PLS_TASK_MANAGER_H_

#include <memory>
#include <utility>
#include <array>
#include <atomic>

#include "pls/internal/scheduling/task.h"
#include "pls/internal/data_structures/stamped_integer.h"
#include "task.h"

namespace pls {
namespace internal {
namespace scheduling {

struct task_handle {
 public:
  enum state { uninitialized, initialized, execute_local, stealing, execute_remote, finished };
  using stamped_state = data_structures::stamped_integer;

  std::atomic<stamped_state> stamped_state_{uninitialized};
  base_task *task_;
};

/**
 * Handles management of tasks in the system. Each thread has a local task manager,
 * responsible for allocating, freeing and publishing tasks for stealing.
 *
 * The manager therefore acts as the deque found in work stealing, as well as the memory
 * management for the tasks (as both are deeply intertwined in our implementation to
 * integrate the memory management into the stealing procedure.
 */
class task_manager {
 public:
  // Publishes a task on the stack, i.e. makes it visible for other threads to steal.
  // The task itself is located on the stack of the worker, as the stealer will copy it away before it is freed.
  void publish_task(base_task &task) {
    task_handle_stack_[tail_internal_].task_ = &task;
    task_handle_stack_[tail_internal_].stamped_state_.store({stamp_internal_++, task_handle::initialized},
                                                            std::memory_order_relaxed);
    tail_internal_++;
    tail_.store(tail_internal_, std::memory_order_release); // Linearization point, handle is published here
  }

  // Try to pop a local task from this task managers stack.
  // This should only be required on the fast path of the implementation,
  // thus only returning if the operation was a success.
  // Essentially this is an 'un-publish' of a task with a notion if it was successful.
  bool steal_local_task() {
    tail_internal_--;
    tail_.store(tail_internal_, std::memory_order_relaxed);

    task_handle::stamped_state swapped_state{task_handle::execute_local, stamp_internal_++};
    task_handle_stack_[tail_internal_].stamped_state_.exchange(swapped_state, std::memory_order_acq_rel);

    if (swapped_state.value == task_handle::execute_remote ||
        swapped_state.value == task_handle::finished) {
      // Someone got the other task, return to 'non linear' execution path
      // TODO: Properly handle slow path
      return false;
    } else {
      // No one got the task so far, we are happy and continue our fast path
      return true;
    }
  }

  // Try to steal a task from a remote task_manager instance. The stolen task must be stored locally.
  // Returns a pair containing the actual task and if the steal was successful.
  // TODO: Re-implement after fast path is done
//  std::pair<task, bool> steal_remote_task(task_manager &other);

  explicit task_manager(task_handle *task_handle_stack) : task_handle_stack_{task_handle_stack},
                                                          head_{{0}},
                                                          tail_{0},
                                                          tail_internal_{0},
                                                          stamp_internal_{0} {}

 private:
  task_handle *task_handle_stack_;
  alignas(base::system_details::CACHE_LINE_SIZE) std::atomic<std::atomic<data_structures::stamped_integer>> head_;
  alignas(base::system_details::CACHE_LINE_SIZE) std::atomic<unsigned int> tail_;
  alignas(base::system_details::CACHE_LINE_SIZE) unsigned int tail_internal_, stamp_internal_;
};

template<size_t NUM_TASKS, size_t MAX_STACK_SIZE>
class static_task_manager {
 public:
  static_task_manager() : static_task_handle_stack_{}, task_manager_{static_task_handle_stack_.data()} {};
  task_manager &get_task_manager() { return task_manager_; }

 private:
  std::array<task_handle, NUM_TASKS> static_task_handle_stack_;
  task_manager task_manager_;
};

}
}
}
#endif //PLS_TASK_MANAGER_H_
